from spotlib._version import __version__ as version
from spotlib.statics import local_config


__author__ = 'Blake Huber'
__version__ = version
__email__ = "blakeca00@gmail.com"


try:

    from libtools import Colors
    from libtools import logd

    # global logger
    logd.local_config = local_config
    logger = logd.getLogger(__version__)

except Exception:
    pass


from spotlib.core import EC2SpotPrices as SpotPrices
from spotlib.core import DurationEndpoints
from spotlib.core import UtcConversion, utc_conversion



"""

lambda_utils (python3)

    Common functionality for use with AWS Lambda Service

Author:
    Blake Huber
    Copyright Blake Huber, All Rights Reserved.

License:
    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee is hereby granted,
    provided that the above copyright notice appear in all copies and that
    both the copyright notice and this permission notice appear in
    supporting documentation

    Additional terms may be found in the complete license agreement:
    https://bitbucket.org/blakeca00/lambda-library-python/src/master/LICENSE.md

"""

import os
import re
import json
import time
import inspect
import logging
import boto3
from botocore.exceptions import ClientError
from spotlib._version import __version__

logger = logging.getLogger(__version__)
logger.setLevel(logging.INFO)


def read_env_variable(arg, default=None, patterns=None):
    """
    Summary:
        Parse environment variables, validate characters, convert
        type(s). default should be used to avoid conversion of an
        variable and retain string type

    Example usage:

    >>> from lambda_utils import read_env_variable
    >>> os.environ['DBUGMODE'] = 'True'
    >>> myvar = read_env_variable('DBUGMODE')
    >>> type(myvar)
    True

    >>> from lambda_utils import read_env_variable
    >>> os.environ['MYVAR'] = '1345'
    >>> myvar = read_env_variable('MYVAR', 'default')
    >>> type(myvar)
    str
    """
    if patterns is None:
        patterns = (
            (re.compile('^[-+]?[0-9]+$'), int),
            (re.compile('\d+\.\d+'), float),
            (re.compile(r'^(true|false)$', flags=re.IGNORECASE), lambda x: x.lower() == 'true'),
            (re.compile('[a-z/]+', flags=re.IGNORECASE), str),
            (re.compile('[a-z/]+\.[a-z/]+', flags=re.IGNORECASE), str),
        )

    if arg in os.environ:
        var = os.environ[arg]
        if var is None:
            ex = KeyError('environment variable %s not set' % arg)
            logger.exception(ex)
            raise ex
        else:
            if default:
                return str(var)     # force default type (str)
            else:
                for pattern, func in patterns:
                    if pattern.match(var):
                        return func(var)
            # type not identified
            logger.warning(
                '%s: failed to identify environment variable [%s] type. May contain \
                special characters' % (inspect.stack()[0][3], arg)
                 )
            return str(var)
    else:
        ex = KeyError('environment variable %s not set' % arg)
        logger.exception(ex)
        raise ex


def get_account_info(account_profile=None):
    """
    Summary:
        Queries AWS iam and sts services to discover account id information
        in the form of account name and account alias (if assigned)

    Returns:
        TYPE: tuple

    Example usage:

    >>> account_number, account_name = lambda_utils.get_account_info()
    >>> print(account_number, account_name)
    103562488773 tooling-prod

    """
    if account_profile:
        session = boto3.Session(profile_name=account_profile)
        sts_client = session.client('sts')
        iam_client = session.client('iam')
    else:
        sts_client = boto3.client('sts')
        iam_client = boto3.client('iam')

    try:
        number = sts_client.get_caller_identity()['Account']
        name = iam_client.list_account_aliases()['AccountAliases'][0]

    except IndexError as e:
        name = '<no_alias_assigned>'
        logger.info('Error: %s. No account alias defined. account_name set to %s' % (e, name))
        return (number, name)
    except ClientError as e:
        logger.warning(
            "%s: problem retrieving caller identity (Code: %s Message: %s)" %
            (inspect.stack()[0][3], e.response['Error']['Code'], e.response['Error']['Message'])
            )
        raise e
    return (number, name)


def get_regions():
    """
    Summary
        Returns list of region codes for all AWS regions worldwide

    Returns:
        TYPE: list

    """
    try:

        client = boto3.client('ec2')
        response = client.describe_regions()

    except ClientError as e:
        logger.critical(
            "%s: problem retrieving aws regions (Code: %s Message: %s)" %
            (inspect.stack()[0][3], e.response['Error']['Code'],
            e.response['Error']['Message']))
        raise e
    return [region['RegionName'] for region in response['Regions']]


def sns_notification(topic_arn, subject, message, account_id=None, account_name=None):
    """
    Summary:
        Sends message to AWS sns service topic provided as a
        parameter

    Args:
        topic_arn (str): sns topic arn
        subject (str): subject of sns message notification
        message (str): message body

    Returns:
        TYPE: Boolean | Success or Failure
    """
    if not (account_id or account_name):
        account_id, account_name = get_account_info()

    # assemble msg
    header = 'AWS Account: %s (%s) | %s' % \
        (str(account_name).upper(), str(account_id), subject)
    msg = '\n%s\n\n%s' % (time.strftime('%c'), message)
    msg_dict = {'default': msg}

    # client
    region = (topic_arn.split('sns:', 1)[1]).split(":", 1)[0]
    client = boto3.client('sns', region_name=region)

    try:
        # sns publish
        response = client.publish(
            TopicArn=topic_arn,
            Subject=header,
            Message=json.dumps(msg_dict),
            MessageStructure='json'
        )
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            return True
        else:
            return False
    except ClientError as e:
        logger.exception(
            '%s: problem sending sns msg (Code: %s Message: %s)' %
            (inspect.stack()[0][3], e.response['Error']['Code'],
                e.response['Error']['Message']))
        return False


def import_file_object(filename):
    """

    Summary: imports block fs object

    Args: block filesystem object

    Returns:
        dictionary obj (valid json file), file data object

    """
    try:
        with open(filename, 'r') as handle:
            file_obj = handle.read()
            dict_obj = json.loads(file_obj)

    except OSError as e:
        logger.critical(
            'import_file_object: %s error opening %s' % (str(e), str(filename))
        )
        raise e
    except ValueError:
        logger.info(
            '%s: import_file_object: %s not json. file object returned' %
            (inspect.stack()[0][3], str(filename))
        )
        return file_obj    # reg file, not valid json
    return dict_obj


def range_bind(min_value, max_value, value):
    """ binds number to a type and range """
    if value not in range(min_value, max_value + 1):
        value = min(value, max_value)
        value = max(min_value, value)
    return int(value)



"""

EC2 SpotPrice Lib, GPL v3 License

Copyright (c) 2018-2019 Blake Huber

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import inspect
import boto3
from botocore.exceptions import ClientError
from spotlib.core import DurationEndpoints
from spotlib.core.utc import utc_conversion
from spotlib.lambda_utils import get_regions
from spotlib import logger


class EC2SpotPrices():
    """
    Generator class using pagination to return unlimited
    number of spot price history data dict

    Methods:
        :set_endpoints (user callable): sets start, end date times for which to request price data
        :_page_iterators: instantiates, constructs a page iterator object
        :_region_paginators (generator): creates regional paginators; one unique per region
        :_spotprice_generator (generator): which uses paginators to request spot price data
        :generate_pricedata (generator, user callable): rollup method for access all child methods

    Use:
        >>>  from spotlib import SpotPrices
        >>>  sp = SpotPrices(page_size=1000)
        >>>  prices = sp.generate_pricedata('us-east-1')

    Returns:
        spot price data (generator)

    """
    def __init__(self, profile='default', start_dt=None, end_dt=None, page_size=500, dt_strings=False, debug=False):
        """
        Args:
            :profile (str): iam identity with appropriate permissions for spot price functionality
            :start_dt (datetime): DateTime object marking data collection start
            :end_dt (datetime): DateTime object marking data collection stop
            :page_size (int):  Number of spot price elements per pagesize
            :dt_strings (bool): if True, return spot price data with isoformat datetime strings
            :debug (bool): debug output toggle
        """
        self.profile = profile
        self.session = boto3.Session(profile_name=self.profile)
        self.regions = get_regions()
        self.start, self.end = self.set_endpoints(start_dt, end_dt)
        self.page_size = page_size
        self.pageconfig = {'PageSize': self.page_size}
        self.dt_strings = dt_strings
        self.debug = debug

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{}(start_dt={}, end_dt={})".format(self.__class__, self.start, self.end)

    def set_endpoints(self, start_dt=None, end_dt=None, duration=None):
        """
        Rationalize start and end datetimes for data history lookup
        """
        self.de = DurationEndpoints()

        if all(x is None for x in [start_dt, end_dt, duration]):
            return self.de.start, self.de.end

        elif duration and start_dt is None:
            s, e = self.de.default_endpoints(duration_days=duration)

        elif start_dt and end_dt:
            s, e = self.de.custom_endpoints(start_time=start_dt, end_time=end_dt)
        self.start, self.end = s, e    # reset instance variable statics
        return s, e

    def _page_iterators(self, region):
        self.client = self.session.client('ec2', region_name=region)
        self.paginator = self.client.get_paginator('describe_spot_price_history')
        self.page_iterator = self.paginator.paginate(
                                StartTime=self.start,
                                EndTime=self.end,
                                DryRun=self.debug,
                                PaginationConfig={'PageSize': self.page_size}
                            )
        return self.page_iterator

    def _region_paginators(self, regions=get_regions()):
        """
        Supplies regional paginator objects, one per unique AWS region
        """
        return [self._page_iterators(region) for region in regions]

    def _spotprice_generator(self, region=None, dt_string=False):
        """
        Generator returning up to 1000 data items per api request to AWS

        Args:
            :region (str): AWS region code. Example: us-east-1
            :dt_string (bool): indicates TYPE for datetime values
                returned in spotprice data.

        Returns:
            spot price data (generator)
        """
        strings = dt_string or self.dt_strings

        for page_iterator in (self._region_paginators() if region is None else self._region_paginators([region])):
            try:

                for page in page_iterator:
                    for price_dict in page['SpotPriceHistory']:
                        yield utc_conversion(price_dict) if strings else price_dict

            except ClientError as e:
                fx = inspect.stack()[0][3]
                logger.exception(f'{fx}: Boto client error while downloading spot history data: {e}')
                continue
            except Exception as e:
                fx = inspect.stack()[0][3]
                logger.exception(f'{fx}: Unknown exception during spot price data retrieval: {e}')

    def generate_pricedata(self, regions, dtstrings=False):
        """
            Rollup facility for ease generation of regional spot price data.
            Iterates child paginator and generator methods to retrieve spot prices.

        Args:
            :region (list): list of AWS region codes (e.g. us-east-1)
            :dtstrings (bool): True returns datetime in str format, DEFAULT: False

        Returns:
            - Spot price data for specific AWS region code
              specified (e.q. region = us-east-1)
        """
        container = []
        for region in regions:
            container.extend([x for x in self._spotprice_generator(region, dtstrings)])
        return {'SpotPriceHistory': container}

    def generate_allregion_pricedata(self, dtstrings=False):
        """
            Rollup facility for ease generation of spot price data from all AWS
            regions. Automates iternation of child paginator and generator methods
            to retrieve spot prices.

        Args:
            :dtstrings (bool): True returns datetime in str format, DEFAULT: False

        Returns:
            - Spot price data for all AWS region codes (e.q. us-east-1)
        """
        return {'SpotPriceHistory': [x for x in self._spotprice_generator(None, dtstrings)]}



"""

Help Menu

    Help menu object containing body of help content.
    For printing with formatting

"""

from spotlib.statics import PACKAGE, PACKAGE_CLI
from spotlib.variables import c, bbc, rst


ACCENT = c.ORANGE               # orange accent highlight color
bdacct = c.ORANGE + c.BOLD      # bold orange
bdcy = c.CYAN + c.BOLD          # bold blue
bdwt = c.BRIGHT_WHITE + c.BOLD          # bold blue
lbrct = bbc + ' [ ' + rst        # left bracket
rbrct = bbc + ' ] ' + rst        # right bracket
vdiv = bbc + ' | ' + rst
tab = '\t'.expandtabs(24)

menu_title = '\n' + bdwt + tab + PACKAGE_CLI + rst + ' command help'

synopsis_cmd = (
        ACCENT + 'spotcli' + rst + ' --start <value> ' + '--end <value>' + vdiv +
        '--duration-days'
    )

url_doc = c.URL + 'http://spotlib.readthedocs.io' + rst
url_sc = c.URL + 'https://github.com/fstab50/spotlib' + rst


menu_body = menu_title + bdwt + """

  DESCRIPTION""" + rst + """

            Command line interface to """ + PACKAGE + """ to retrieve Amazon
            EC2 Spot Instance Pricing and associated metadata.

            Source Code Repo: """ + url_sc + """
    """ + bdwt + """
  SYNOPSIS""" + rst + """

        $ """ + synopsis_cmd + """

                        -r, --region <value>
                       [-s, --start  <value>  ]
                       [-e, --end    <value>  ]
                       [-l, --list   <methods> | <regions>  ]
                       [-d, --debug    ]
                       [-h, --help     ]
                       [-V, --version  ]
    """ + bdwt + """
  OPTIONS
    """ + bdwt + """
        -D, --duration-days""" + rst + """ <value>: Number of days of price data
            history to retrieve ending at midnight on present day
    """ + bdwt + """
        -e, --end""" + rst + """ <value>:  Datetime of end of the price sampling
            period (example: 2019-09-04T23:59:59). See --start.
    """ + bdwt + """
        -l, --list""" + rst + """ <value>: List available method calls or Amazon
            Web Services' region codes
    """ + bdwt + """
        -d, --debug""" + rst + """:  Print out additional  debugging information
    """ + bdwt + """
        -h, --help""" + rst + """: Show this help message, symbol legend, & exit
    """ + bdwt + """
        -r, --region""" + rst + """:  AWS region code (e.g. us-east-1) for which
            you wish to retrieve EC2 spot price data
    """ + bdwt + """
        -s, --start""" + rst + """ <value>:  Datetime of start of price sampling
            period (example: 2019-09-03T00:00:00). See --end.
    """ + bdwt + """
        -V, --version""" + rst + """: Print version, license, and copyright info
    """



from spotlib._version import __version__ as version


__author__ = 'Blake Huber'
__version__ = version
__email__ = "blakeca00@gmail.com"


from spotlib.core.endpoints import DurationEndpoints
from spotlib.core.spotcore import EC2SpotPrices
from spotlib.core.utc import UtcConversion, utc_conversion



"""
Summary.

    Python 3 Class.  Datatime object conversion --> utc formatted string

"""

import datetime


def utc_conversion(data):
    """
        Converts datetime object embedded in a dictionary schema
        to utc datetime string format

    Args:
        :data (list | dict):  list of spot price data.  Alternatively,
            can be same list wrapped in dictionary (below)

    .. code: json

        {
            'AvailabilityZone': 'eu-west-1a',
            'InstanceType': 'm5d.4xlarge',
            'ProductDescription': 'Red Hat Enterprise Linux',
            'SpotPrice': '0.420000',
            'Timestamp': datetime.datetime(2019, 8, 11, 23, 56, 50, tzinfo=tzutc())
        }

    Returns:
        datetime, TYPE: str

    .. code: json

        {
            'AvailabilityZone': 'eu-west-1a',
            'InstanceType': 'm5d.4xlarge',
            'ProductDescription': 'Red Hat Enterprise Linux',
            'SpotPrice': '0.420000',
            'Timestamp': '2019-08-11T23:56:50Z'
        }

    """
    dt = data['Timestamp'].strftime('%Y-%m-%dT%H:%M:%SZ')
    data['Timestamp'] = dt
    return data


class UtcConversion():
    """
    Class that converts datetime object embedded in a
    dictionary schema to utc datetime string format
    """
    def __init__(self, data):
        """
        Args:
            :data (list | dict):  list of spot price data.  Alternatively,
                can be same list wrapped in dictionary (below)

        .. code: json

            {
                'SpotPriceHistory': [
                    {
                        'AvailabilityZone': 'eu-west-1a',
                        'InstanceType': 'm5d.4xlarge',
                        'ProductDescription': 'Red Hat Enterprise Linux',
                        'SpotPrice': '0.420000',
                        'Timestamp': datetime.datetime(2019, 8, 11, 23, 56, 50, tzinfo=tzutc())
                    },
                    {
                        'AvailabilityZone': 'eu-west-1a',
                        'InstanceType': 'm5d.4xlarge',
                        ...
                    }
                ]
            }
        """
        self.d = data['SpotPriceHistory'] if isinstance(data, dict) else data
        self.formatted = [x['Timestamp'].strftime('%Y-%m-%dT%H:%M:%SZ') for x in self.d]
        self.prices = self.convert(self.d)

    def convert(self, pricelist):
        """
        Converts datetime object embedded in a dictionary
        schema to utc datetime string format

        Args:
            :pricelist (list): list of spot price dictionaries
                with time represented as a datetime object

        Returns:
            list of dictionaries, TYPE: list

        .. code: json

            {
                'AvailabilityZone': 'eu-west-1a',
                'InstanceType': 'm5d.4xlarge',
                'ProductDescription': 'Red Hat Enterprise Linux',
                'SpotPrice': '0.420000',
                'Timestamp': '2019-08-11T23:56:50Z'
            }

        """
        for index, pdict in enumerate(pricelist):
            pricelist[index]['Timestamp'] = self.formatted[index]
        return pricelist



"""

ec2 spotprice retriever, GPL v3 License

Copyright (c) 2018-2019 Blake Huber

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import os
import sys
import datetime
import json
import inspect
import argparse
import subprocess
import boto3
from botocore.exceptions import ClientError
from libtools import stdout_message
from libtools.js import export_iterobject
from spotlib import SpotPrices, UtcConversion
from spotlib.help_menu import menu_body
from spotlib import about, logger
from spotlib.variables import acct, bdwt, bbc, bbl, bcy, btext, rst


try:
    from libtools.oscodes_unix import exit_codes
    os_type = 'Linux'
    user_home = os.getenv('HOME')
    splitchar = '/'                                   # character for splitting paths (linux)

except Exception:
    from libtools.oscodes_win import exit_codes         # non-specific os-safe codes
    os_type = 'Windows'
    user_home = os.getenv('username')
    splitchar = '\\'                                  # character for splitting paths (windows)


# globals
container = []
module = os.path.basename(__file__)
iloc = os.path.abspath(os.path.dirname(__file__))     # installed location of modules


def _debug_output(*args):
    """additional verbose information output"""
    for arg in args:
        if os.path.isfile(arg):
            print('Filename {}'.format(arg.strip(), 'lower'))
        elif str(arg):
            print('String {} = {}'.format(getattr(arg.strip(), 'title'), arg))


def default_endpoints(duration_days=1):
    """
    Supplies the default start and end datetime objects in absence
    of user supplied endpoints which frames time period from which
    to begin and end retrieving spot price data from Amazon APIs.

    Returns:  TYPE: tuple, containing:
        - start (datetime), midnight yesterday
        - end (datetime) midnight, current day

    """
    # end datetime calcs
    dt_date = datetime.datetime.today().date()
    dt_time = datetime.datetime.min.time()
    end = datetime.datetime.combine(dt_date, dt_time)

    # start datetime calcs
    duration = datetime.timedelta(days=duration_days)
    start = end - duration
    return start, end


def format_pricefile(key):
    """Adds path delimiter and color formatting to output artifacts"""
    region = bcy + key.split('/')[0] + rst
    pricefile = bcy + key.split('/')[1] + rst
    delimiter = bdwt + '/' + rst
    return region + delimiter + pricefile


def help_menu():
    """Print help menu options"""
    print(menu_body)


def local_awsregion(profile):
    """Determines AWS region code local to user"""
    if os.environ.get('AWS_DEFAULT_REGION'):
        return os.environ['AWS_DEFAULT_REGION']
    cmd = 'aws configure get {}.region'.format(profile)
    return subprocess.getoutput(cmd).strip()


def summary_statistics(data, instances):
    """
    Calculate stats across spot price data elements retrieved
    in the current execution.  Prints to stdout

    Args:
        :data (list): list of spot price dictionaries
        :instances (list): list of unique instance types found in data

    Returns:
        Success | Failure, TYPE:  bool
    """
    instance_dict, container = {}, []

    for itype in instances:
        try:
            cur_type = [
                x['SpotPrice'] for x in data['SpotPriceHistory'] if x['InstanceType'] == itype
            ]
        except KeyError as e:
            logger.exception('KeyError on key {} while printing summary report statistics.'.format(e))
            continue

        instance_dict['InstanceType'] = str(itype)
        instance_dict['AvgPrice'] = sum([float(x['SpotPrice']) for x in cur_type]) / len(cur_type)
        container.append(instance_dict)
    # output to stdout
    print_ending_summary(instances, container)
    return True


def print_ending_summary(itypes_list, summary_data):
    """
    Prints summary statics to stdout at the conclusion of spot
    price data retrieval
    """
    now = datetime.datetime.now().strftime('%Y-%d-%m %H:%M:%S')
    tab = '\t'.expandtabs(4)
    print('EC2 Spot price data retrieval concluded {}'.format(now))
    print('Found {} unique EC2 size types in spot data'.format(len(itypes_list)))
    print('Instance Type distribution:')
    for itype in itypes_list:
        for instance in container:
            if instance['InstanceType'] == itype:
                print('{} - {}'.format(tab, itype, instance['AvgPrice']))


def source_environment(env_variable):
    """
    Sources all environment variables
    """
    return {
        'duration_days': read_env_variable('default_duration'),
        'page_size': read_env_variable('page_size', 500),
        'bucket': read_env_variable('S3_BUCKET', None)
    }.get(env_variable, None)


def modules_location():
    """Filsystem location of Python3 modules"""
    return os.path.split(os.path.abspath(__file__))[0]


def options(parser, help_menu=False):
    """
    Summary:
        parse cli parameter options

    Returns:
        TYPE: argparse object, parser argument set

    """
    # default datetime objects when no custom datetimes supplied
    start_dt, end_dt = default_endpoints()

    parser.add_argument("-C", "--configure", dest='configure', action='store_true', required=False)
    parser.add_argument("-d", "--debug", dest='debug', action='store_true', default=False, required=False)
    parser.add_argument("-e", "--end", dest='end', nargs='*', default=end_dt, required=False)
    parser.add_argument("-h", "--help", dest='help', action='store_true', required=False)
    parser.add_argument("-p", "--profile", dest='profile', nargs=1, default='default', required=False)
    parser.add_argument("-r", "--region", dest='region', nargs='*', default='noregion', required=False)
    parser.add_argument("-D", "--duration-days", dest='duration', nargs='*', default=None, required=False)
    parser.add_argument("-s", "--start", dest='start', nargs='*', default=start_dt, required=False)
    parser.add_argument("-V", "--version", dest='version', action='store_true', required=False)
    return parser.parse_known_args()


def package_version():
    """
    Prints package version and requisite PACKAGE info
    """
    print(about.about_object)
    sys.exit(exit_codes['EX_OK']['Code'])


def precheck(debug, region):
    """
    Runtime Dependency Checks: postinstall artifacts, environment
    """
    if region == 'noregion':
        return False
    return True

    try:

        home_dir = os.expanduser('~')
        config_file = os.path.join(home_dir, '.spotlib.json')

        if os.path.exists(config_file):
            with open(config_file, 'r') as f1:
                defaults = json.loads(f1.read())
        else:
            from spotlib.defaults import defaults

        _debug_output(home_dir, config_file)

    except OSError:
        fx = inspect.stack()[0][3]
        logger.exception('{}: Problem installing user config files. Exit'.format(fx))
        return False
    return defaults


def s3upload(bucket, s3object, key, profile='default'):
    """
        Streams object to S3 for long-term storage

    Returns:
        Success | Failure, TYPE: bool
    """
    try:
        session = boto3.Session(profile_name=profile)
        s3client = session.client('s3')
        # dict --> str -->  bytes (utf-8 encoded)
        bcontainer = json.dumps(s3object, indent=4, default=str).encode('utf-8')
        response = s3client.put_object(Bucket=bucket, Body=bcontainer, Key=key)

        # http completion code
        statuscode = response['ResponseMetadata']['HTTPStatusCode']

    except ClientError as e:
        logger.exception(f'Unknown exception while calc start & end duration: {e}')
        return False
    return True if str(statuscode).startswith('20') else False


def writeout_data(key, jsonobject, filename):
    """
        Persists json data to local filesystem

    Returns:
        Success | Failure, TYPE: bool

    """
    tab = '\t'.expandtabs(13)

    if export_iterobject({key: jsonobject}, filename):
        success = f'Wrote {bcy + filename + rst}\n{tab}successfully to local filesystem'
        stdout_message(success, prefix='OK')
        return True
    else:
        failure = f'Problem writing {bcy + filename + rst} to local filesystem'
        stdout_message(failure, prefix='WARN')
        return False



def init():
    """
    Initialize spot price operations; process command line parameters
    """
    parser = argparse.ArgumentParser(add_help=False)

    try:

        args, unknown = options(parser)

    except Exception as e:
        help_menu()
        stdout_message(str(e), 'ERROR')
        sys.exit(exit_codes['E_BADARG']['Code'])

    if len(sys.argv) == 1 or args.help:
        help_menu()
        sys.exit(exit_codes['EX_OK']['Code'])

    elif args.version:
        package_version()

    elif (args.start and args.end) or args.duration:
        # set local region
        args.region = local_awsregion(args.profile) if args.region == 'noregion' else args.region

        # validate prerun conditions
        defaults = precheck(args.debug, args.region)

        sp = SpotPrices(profile=args.profile)

        if args.duration and isinstance(int(args.duration[0]), int):
            start, end = sp.set_endpoints(duration=int(args.duration[0]))
        else:
            start, end = sp.set_endpoints(args.start, args.end)

        # global container for ec2 instance size types
        instance_sizes = []

        for region in args.region:

            fname = '_'.join(
                        [
                            start.strftime('%Y-%m-%dT%H:%M:%SZ'),
                            end.strftime('%Y-%m-%dT%H:%M:%SZ'),
                            'all-instance-spot-prices.json'
                        ]
                    )

            prices = sp.generate_pricedata(regions=[region])

            # conversion of datetime obj => utc strings
            uc = UtcConversion(prices)

            # write to file on local filesystem
            key = os.path.join(region, fname)
            os.makedirs(region) if not os.path.exists(region) else True
            _completed = export_iterobject(prices, key)

            # log status
            tab = '\t'.expandtabs(13)
            fkey = format_pricefile(key)
            success = f'Wrote {fkey}\n{tab}successfully to local filesystem'
            failure = f'Problem writing {fkey} to local filesystem'
            stdout_message(success, prefix='OK') if _completed else stdout_message(failure, prefix='WARN')

            # build unique collection of instances for this region
            regional_sizes = list(set([x['InstanceType'] for x in prices['SpotPriceHistory']]))
            instance_sizes.extend(regional_sizes)

        # instance sizes across analyzed regions
        instance_sizes = list(set(instance_sizes))
        instance_sizes.sort()
        key = 'instanceTypes'
        date = sp.end.strftime("%Y-%m-%d")
        return writeout_data(key, instance_sizes, date + '_spot-instanceTypes.json')

    else:
        stdout_message(
            'Dependency check fail %s' % json.dumps(args, indent=4),
            prefix='AUTH',
            severity='WARNING'
            )
        sys.exit(exit_codes['E_DEPENDENCY']['Code'])

    failure = """ : Check of runtime parameters failed for unknown reason.
    Please ensure you have both read and write access to local filesystem. """
    logger.warning(failure + 'Exit. Code: %s' % sys.exit(exit_codes['E_MISC']['Code']))
    print(failure)
    return sys.exit(exit_codes['E_BADARG']['Code'])



"""

EC2 SpotPrice Lib, GPL v3 License

Copyright (c) 2018-2019 Blake Huber

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import os
import sys
import re
import datetime
import inspect
from spotlib import logger


# precompiled regex pattern; datetime
re_dt = re.compile('\d{4}-[01]\d-[0-3]\d[\sT][0-2]\d:[0-5]\d:[0-5]\d(?:\.\d+)?Z?')

# precompiled regex pattern; No timezone information -- NOT VALID SYNTAX
re_notz = re.compile('\d{4}-[01]\d-[0-3]\d[\sT][0-2]\d:[0-5]\d:[0-5]\d(?:\.\d+)^Z')

# precompiled regex pattern; datetime, no "T" separator
re_dtnot = re.compile('\d{4}-[01]\d-[0-3]\d[\s][0-2]\d:[0-5]\d:[0-5]\d(?:\.\d+)?Z?')

# precompiled regex pattern; date
re_date = re.compile('\d{4}-[01]\d-[0-3]\d')


def format_datetime(datetime_str):
    """
    Helper module function: Formats datetime strings & dates
    (datetime with no time component)

    Returns:
        datetime iso formatted string
    """
    def convert_dt(datetime_str):
        dt = datetime.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')
        return dt.isoformat()

    if isinstance(datetime_str, str) and re_dt.match(datetime_str):
        return datetime_str if not re_dtnot.match(datetime_str) else convert_dt(datetime_str)
    return ''.join([datetime_str, 'T00:00:00']) if re_date.match(datetime_str) else datetime_str


class DurationEndpoints():
    """
    Calculates both custom and default endpoints in time which brackets
    the time period for which spot price historical data is retrieved

    Methods:
        :default_endpoints: construct default start, end datetimes;
            i.e. midnight previous day to midnight current day
        :custom_endpoints: construct customised start, end datetimes via endpoints
        :custom_duration: construct customised start, end datetimes via duration only
        :_convert_dt_string: method to reliably transform various datetime inputs into dt strings

    Use:
        from spotlib import DurationEndpoints
        d = DurationEndpoints()
    """
    def __init__(self, duration_days=1, start_dt=None, end_dt=None, debug=False):
        """
        Args:
            :duration_days (int): Number of days between endpoint dates
            :start_dt (datetime): Datetime start of spot price retrieval period
            :end_dt (datetime): Datetime endpoint of spot price retrieval period
            :debug (bool): verbose output flag

        """
        self.d_days = duration_days

        if all(x is None for x in [start_dt, end_dt]):
            self.start, self.end = self.default_endpoints(self.d_days)

        elif any(x is not None for x in [start_dt, end_dt]):
            x, y = self.custom_endpoints(start_dt, end_dt)
            self.start = x if x is not None else self.default_endpoints()[0]
            self.end = y if y is not None else self.default_endpoints()[1]

    def default_endpoints(self, duration_days=1):
        """
        Supplies the default start and end dates (dt objects) in absence
        of user supplied endpoints which frames time period from which
        to begin and end retrieving spot price data from Amazon APIs.

        Returns:  TYPE: tuple, containing:

            - start (datetime), midnight yesterday
            - end (datetime) midnight, current day

        """
        # end datetime calcs
        dt_date = datetime.datetime.today().date()
        dt_time = datetime.datetime.min.time()
        end = datetime.datetime.combine(dt_date, dt_time)

        # start datetime calcs
        duration = datetime.timedelta(days=duration_days)
        start = end - duration
        return start, end

    def custom_endpoints(self, start_time=None, end_time=None):
        """
        Calculates custom start and end dates when given a variety of
        formats including string or None. If both duration_days and start_time,
        end_time values are provided, start and end times will take precedence.

        Args:
            :duration_days (int): Duration between start and end points in 24h days
            :start_time (datetime | str | None):  midnight on provided custom date
            :end_time (datetime | str | None):  midnight on provided custom date

        Returns:
            start, end: points in time, TYPE:  datetime regardless of input format

        """
        try:
            if all(isinstance(x, datetime.datetime) for x in [start_time, end_time]):
                return start_time, end_time

            elif any(isinstance(x, str) for x in [start_time, end_time]) \
                and (re_dt.match(x) for x in [start_time, end_time]):
                start = self._convert_dt_string(start_time)
                end = self._convert_dt_string(end_time)

            elif any(x is None for x in [start_time, end_time]):
                start, end = self.default_endpoints()
        except Exception as e:
            fx = inspect.stack()[0][3]
            logger.exception(f'{fx}: Unknown exception while calc start & end duration: {e}')
            return self.start, self.end
        return start, end

    def custom_duration(self, duration_days):
        """
        Returns start and end datetimes when given a custom duration in days
        """
        return self.default_endpoints(duration_days)

    def _convert_dt_string(self, dt_str):
        dt_format = '%Y-%m-%dT%H:%M:%SZ'
        return datetime.datetime.strptime(format_datetime(dt_str), dt_format)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{}(start_dt={}, end_dt={})".format(self.__class__, self.start, self.end)



"""
Default runtime parameters:

    - Used for execution outside of AWS lambda
    - Substitue for Lambda environment variables
    - Run spotcli command line utility to create
      configuration file

    .. code: bash

        $ spotcli --configure

"""

defaults = {
    "iam_profile": "default",
    "s3_bucket": "mybucket",
    "duration_days": "1"
}



"""

spotlib :  Copyright 2018-2019, Blake Huber

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

see: https://www.gnu.org/licenses/#GPL

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
contained in the program LICENSE file.

"""

import os
import sys
import platform
import subprocess
from shutil import which
from setuptools import setup, find_packages
import getpass
from codecs import open
import spotlib


requires = [
    'boto3>=1.9.100',
    'libtools>=0.2.4'
]


_project = 'spotlib'
_root = os.path.abspath(os.path.dirname(__file__))


def _root_user():
    """
    Checks localhost root or sudo access.  Returns bool
    """
    if os.geteuid() == 0:
        return True
    elif subprocess.getoutput('echo $EUID') == '0':
        return True
    return False


def _user():
    """Returns username of caller"""
    return getpass.getuser()


def _set_pythonpath():
    """
    Temporarily reset PYTHONPATH to prevent home dir = python module home
    """
    os.environ['PYTHONPATH'] = '/'


def module_dir():
    """Filsystem location of Python3 modules"""
    bin_path = which('python3.6') or which('python3.7')
    bin = bin_path.split('/')[-1]
    if 'local' in bin:
        return '/usr/local/lib/' + bin + '/site-packages'
    return '/usr/lib/' + bin + '/site-packages'


def os_parityPath(path):
    """
    Converts unix paths to correct windows equivalents.
    Unix native paths remain unchanged (no effect)
    """
    path = os.path.normpath(os.path.expanduser(path))
    if path.startswith('\\'):
        return 'C:' + path
    return path


def preclean(dst):
    if os.path.exists(dst):
        os.remove(dst)
    return True


def read(fname):
    basedir = os.path.dirname(sys.argv[0])
    return open(os.path.join(basedir, fname)).read()


def user_home():
    """Returns os specific home dir for current user"""
    try:
        if platform.system() == 'Linux':
            return os.path.expanduser('~')

        elif platform.system() == 'Windows':
            username = os.getenv('username')
            return 'C:\\Users\\' + username

        elif platform.system() == 'Java':
            print('Unable to determine home dir, unsupported os type')
            sys.exit(1)
    except OSError as e:
        raise e


setup(
    name=_project,
    version=spotlib.__version__,
    description='Library for retrieving Amazon EC2 Spot Price Data',
    long_description=read('DESCRIPTION.rst'),
    url='https://github.com/fstab50/spotlib',
    author=spotlib.__author__,
    author_email=spotlib.__email__,
    license='GPL-3.0',
    classifiers=[
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
    ],
    keywords='Amazon AWS EC2 spot prices lambda reports cost management',
    packages=find_packages(exclude=['assets', 'docs', 'reports', 'scripts', 'tests']),
    install_requires=requires,
    python_requires='>=3.6, <4',
    entry_points={
        'console_scripts': [
            'spotcli=spotlib.cli:init'
        ]
    },
    zip_safe=False
)
